// CAWireWorld2DRules - Rules for WireWorld
// Implementation file
/////////////////////////////////////////////////////////////////////////////

// CADLib includes
#include <AfxDpnds.h>
#include <CADLib.h>

// Component dependent includes
#include "Resource.h"
#include "WireWorld2D.Zones.h"
#include "CompsSrc\\Data\\CrtsInt2D\\CrtsInt2D.h"

// Implements the autentification function
COMPATIBLE_RULES(1.1)

// Implements the creation/destruction functions
RULES_COMPONENT(CAWireWorld2DRules)

CAWireWorld2DRules::CAWireWorld2DRules():
	oOpt(1, 1),
	p_iElectrons("Electrons", "Amount of Electron Head cells", 0)
{
	sComment = "WireWorld";
	bFinalizeIfChanged = true;
}

bool CAWireWorld2DRules::Initialize()
{
	// Prepare optimizer
	oOpt.Reset();

	// Now variable "datum" points to a typecasted datum component
	DATUM(CACrtsInt2DDatum);
	CACell c; // Current cell

	iq = 0; // Electron heads counter
	
	// Loop over all cells to count "electron head" cells
	for (CACell i = datum->z.a1; i <= datum->z.b1; i++)
	{
		for (CACell j = datum->z.a2; j <= datum->z.b2; j++)
		{
			// Current cell
			c = GetUnion()->Metrics->ToCell(i, j, 0);
			
			// Build up electron count
			if (datum->Get(c) == WW_CELL_HEAD) iq++;
			
			// Initialize optimizer:
			// we are only interested non-background zones
			// as any activity could occur only at them
			if (datum->Get(c) != WW_CELL_BACK)
			{
				oOpt.AddCell(i, j);
			}
		}
	}
	
	// Update electron heads counter
	p_iElectrons.Set(iq);

	// Initialize double buffer for optimizer to work properly
	if (datum->p_bDouble.Get())
	{
		for (CACell i = datum->z.a1; i <= datum->z.b1; i++)
		{
			for (CACell j = datum->z.a2; j <= datum->z.b2; j++)
			{
				c = GetUnion()->Metrics->ToCell(i, j, 0);
				datum->Set(c, WW_CELL_BACK);
			}
		}
	}

	return true;
}


bool CAWireWorld2DRules::SubCompute(Zone&)
{
	// Now variable "datum" points to a typecasted datum component
	DATUM(CACrtsInt2DDatum);

	CACell c;                // Current cell
	CACell neig[8];          // All neighbors in the definite metrics
	unsigned int ncount = GET_METRICS->GetNeighboursCount(); // Stores neighbors count
	unsigned char ntypes[4]; // neighbor counts for each of 4 cell states
	iq = 0;                  // Electron heads counter

	
	// Prepare optimizer
	oOpt.NewStep();
	Zone z;

	// Iterate over all zones marked as "interesting" at previous step
	while (oOpt.ReleaseZone(z))
	{
		// Loop over all cells in zone to be recalculated
		for (CACell i = z.a1; i <= z.b1; i++)
		{
			for (CACell j = z.a2; j <= z.b2; j++)
			{
				// Initialize neighbor counts
				for (unsigned int k = 0; k < 4; k++) ntypes[k] = 0;

				// Current cell
				c = GetUnion()->Metrics->ToCell(i, j, 0);

				// Retrieve all neighbors
				pUnion->Metrics->GetNeighbours(c, neig);

				// Calculate neighbor counts of all states
				for (k = 0; k < ncount; k++)
				{
					ntypes[datum->Get(neig[k])]++;
				}

				// Determine next state of current cell according to WireWorld rules
				int new_state;
				switch (datum->Get(c))
				{
					case WW_CELL_BACK:
						// Background always remain background
						new_state = WW_CELL_BACK;
						break;

					case WW_CELL_WIRE:
						// Wire becomes electron head if and only if
						// it has 1 or 2 "electron head" neighbors
						new_state = (ntypes[WW_CELL_HEAD] == 1 || ntypes[WW_CELL_HEAD] == 2)
							? WW_CELL_HEAD
							: WW_CELL_WIRE;

						// Adjust electron heads counter
						if (new_state == WW_CELL_HEAD) iq++;
						break;

					case WW_CELL_HEAD:
						// Electron head always change to tail
						new_state = WW_CELL_TAIL;
						break;

					case WW_CELL_TAIL:
						// Electron tail always change to wire
						new_state = WW_CELL_WIRE;
						break;

					default:
						// Everything weird is eliminated (should not occur normally)
						new_state = WW_CELL_BACK;
				}

				// If state is changed, add it to "interesting cells" set for next step
				if (new_state != datum->Get(c))
				{
					oOpt.AddCell(i, j);
				}

				// Update current cell state
				datum->Set(c, new_state);
			}
		}
	}
	p_iElectrons.Set(iq);

	return true;
}
